@@ -439,6 +439,9 @@ class DataValidator:
             raise ValueError(msg)
 
     def check_flags(self, flags: pl.DataFrame, time_sku_data: pl.DataFrame) -> None:
+        if flags.is_empty():
+            logger.info("Flags DataFrame is empty, skipping validation.")
+            return
         self.validate_schema(flags, flags_schema, "flags")
 
         # Check that all flag values are valid according to FlagLevels enum using Polars native functions
@@ -825,6 +828,9 @@ class DataValidator:
         # Validate continuous feature values
         self._validate_continuous_feature_values(description, data)
 
+        # Validate text feature values
+        self._validate_text_feature_values(description, data)
+
         # Validate date range if data has date column
         if "date" in data.columns:
             dates = data["date"].unique().sort()
@@ -843,18 +849,39 @@ class DataValidator:
 
         if categorical_features.is_empty():
             return  # No categorical features to validate
+        
+        # Create allowed values DataFrame
+        levels_df = (
+            categorical_features
+            .select("name", "levels")
+            .explode("levels")
+            .rename({"name": "feature", "levels": "allowed_value"})
+        )
 
-        # Join data with categorical features and validate in one operation
+        # Get only categorical feature data by joining with categorical_features
+        categorical_data = data.join(
+            categorical_features.select("name"), 
+            left_on="feature", 
+            right_on="name", 
+            how="inner"
+        ).filter(pl.col("value").is_not_null())
+
+        if categorical_data.is_empty():
+            return
+
+        # Find invalid values using left join and check for nulls
         invalid_data = (
-            data.join(categorical_features, left_on="feature", right_on="name", how="left")
-            .filter(pl.col("var_type").is_not_null())  # Only categorical features
-            .filter(pl.col("value").is_not_null() & ~pl.col("levels").list.contains(pl.col("value")))
+            categorical_data
+            .join(levels_df, left_on=["feature", "value"], right_on=["feature", "allowed_value"], how="left")
+            .filter(pl.col("allowed_value").is_null())  # Invalid values will have null allowed_value
         )
 
         if not invalid_data.is_empty():
-            # Get sample of invalid values per feature for error message (limit to 10)
+            # Join back with categorical_features to get the levels for error message
             invalid_summary = (
-                invalid_data.group_by(["feature", "levels"])
+                invalid_data
+                .join(categorical_features.select(["name", "levels"]), left_on="feature", right_on="name")
+                .group_by(["feature", "levels"])
                 .agg(pl.col("value").unique().head(10).alias("invalid_values"))
                 .sort("feature")
             )
@@ -864,12 +891,24 @@ class DataValidator:
                 feature_name = row["feature"]
                 allowed_levels = row["levels"]
                 invalid_values = row["invalid_values"]
+                # Truncate long level lists for readability
+                levels_display = allowed_levels[:10] if len(allowed_levels) > 10 else allowed_levels
+                levels_suffix = "..." if len(allowed_levels) > 10 else ""
                 error_details.append(
-                    f"  - {feature_name}: invalid values {invalid_values}, allowed: {allowed_levels} (plus null)"
+                    f"  - {feature_name}: invalid values {invalid_values}, allowed: {levels_display}{levels_suffix} (plus null)"
                 )
 
             msg = "Categorical features have invalid values:\n" + "\n".join(error_details)
             raise ValueError(msg)
+    
+    def _validate_text_feature_values(self, description: pl.DataFrame, data: pl.DataFrame) -> None:
+        """Validate that text feature values are strings or null."""
+
+        # check that columns "value" in data is Utf8
+        if data.select(pl.col("value").dtype != pl.Utf8):
+            raise TypeError("Text feature values must be of type Utf8")
+    
+
 
     def _validate_binary_feature_values(self, description: pl.DataFrame, data: pl.DataFrame) -> None:
         """Validate that binary feature values are only 0 or 1."""
